
import torch
from abc import ABC
from typing import Tuple


class BaseSampler(ABC):
    r"""An abstract base class that initializes a graph sampler and provides
    :meth:`_sample_one_layer_from_nodes`
    :meth:`_sample_one_layer_from_nodes_parallel`
    :meth:`sample_from_nodes` routines.
    """

    def sample_from_nodes(
        self,
        nodes: torch.Tensor,
        **kwargs
    ) -> Tuple[torch.Tensor, list]:
        r"""Performs mutilayer sampling from the nodes specified in: nodes
        The specific number of layers is determined by parameter: num_layers
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, list].

        Args:
            nodes: the list of seed nodes index
            **kwargs: other kwargs
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index: the edges sampled
        """
        raise NotImplementedError

    def _sample_one_layer_from_nodes(
        self,
        nodes:torch.Tensor,
        **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Performs sampling from the nodes specified in: nodes,
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, torch.Tensor].

        Args:
            nodes: the list of seed nodes index
            **kwargs: other kwargs
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index: the edges sampled
        """
        raise NotImplementedError
    
    def _sample_one_layer_from_nodes_parallel(
        self, 
        nodes: torch.Tensor,
        **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Performs sampling paralleled from the nodes specified in: nodes,
        returning a sampled subgraph in the specified output format: Tuple[torch.Tensor, torch.Tensor].

        Args:
            nodes: the list of seed nodes index
            **kwargs: other kwargs
        Returns:
            sampled_nodes: the nodes sampled
            sampled_edge_index: the edges sampled
        """
        raise NotImplementedError
